@@ -3,12 +3,6 @@
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
-// Defined by build system; this avoids IDE warnings. Must come before
-// color_management.h (affects header definitions).
-#ifndef JPEGXL_ENABLE_SKCMS
-#define JPEGXL_ENABLE_SKCMS 0
-#endif
-
 #include "lib/jxl/color_management.h"
 
 #include <math.h>
@@ -33,6 +27,7 @@
 #include "lib/jxl/base/status.h"
 #include "lib/jxl/field_encodings.h"
 #include "lib/jxl/linalg.h"  // MatMul, Inv3x3Matrix
+#include "lib/jxl/opsin_params.h"
 #include "lib/jxl/transfer_functions-inl.h"
 
 HWY_BEFORE_NAMESPACE();
@@ -219,6 +214,11 @@ void WriteICCUint16(uint16_t value, size_t pos, PaddedBytes* JXL_RESTRICT icc) {
   (*icc)[pos + 1] = value & 255;
 }
 
+void WriteICCUint8(uint8_t value, size_t pos, PaddedBytes* JXL_RESTRICT icc) {
+  if (icc->size() < pos + 1) icc->resize(pos + 1);
+  (*icc)[pos] = value;
+}
+
 // Writes a 4-character tag
 void WriteICCTag(const char* value, size_t pos, PaddedBytes* JXL_RESTRICT icc) {
   if (icc->size() < pos + 4) icc->resize(pos + 4);
@@ -249,8 +249,10 @@ Status CreateICCHeader(const ColorEncoding& c,
 
   WriteICCUint32(0, 0, header);  // size, correct value filled in at end
   WriteICCTag(kCmm, 4, header);
-  WriteICCUint32(0x04300000u, 8, header);
-  WriteICCTag("mntr", 12, header);
+  WriteICCUint32(0x04400000u, 8, header);
+  const char* profile_type =
+      c.GetColorSpace() == ColorSpace::kXYB ? "scnr" : "mntr";
+  WriteICCTag(profile_type, 12, header);
   WriteICCTag(c.IsGray() ? "GRAY" : "RGB ", 16, header);
   WriteICCTag("XYZ ", 20, header);
 
@@ -337,6 +339,44 @@ Status CreateICCChadTag(float chad[9], PaddedBytes* JXL_RESTRICT tags) {
   return true;
 }
 
+void MaybeCreateICCCICPTag(const ColorEncoding& c,
+                           PaddedBytes* JXL_RESTRICT tags, size_t* offset,
+                           size_t* size, PaddedBytes* JXL_RESTRICT tagtable,
+                           std::vector<size_t>* offsets) {
+  if (c.GetColorSpace() != ColorSpace::kRGB) {
+    return;
+  }
+  uint8_t primaries = 0;
+  if (c.primaries == Primaries::kP3) {
+    if (c.white_point == WhitePoint::kD65) {
+      primaries = 12;
+    } else if (c.white_point == WhitePoint::kDCI) {
+      primaries = 11;
+    } else {
+      return;
+    }
+  } else if (c.primaries != Primaries::kCustom &&
+             c.white_point == WhitePoint::kD65) {
+    primaries = static_cast<uint8_t>(c.primaries);
+  } else {
+    return;
+  }
+  if (c.tf.IsUnknown() || c.tf.IsGamma()) {
+    return;
+  }
+  WriteICCTag("cicp", tags->size(), tags);
+  WriteICCUint32(0, tags->size(), tags);
+  WriteICCUint8(primaries, tags->size(), tags);
+  WriteICCUint8(static_cast<uint8_t>(c.tf.GetTransferFunction()), tags->size(),
+                tags);
+  // Matrix
+  WriteICCUint8(0, tags->size(), tags);
+  // Full range
+  WriteICCUint8(1, tags->size(), tags);
+  FinalizeICCTag(tags, offset, size);
+  AddToICCTagTable("cicp", *offset, *size, tagtable, offsets);
+}
+
 void CreateICCCurvCurvTag(const std::vector<uint16_t>& curve,
                           PaddedBytes* JXL_RESTRICT tags) {
   size_t pos = tags->size();
@@ -349,6 +389,7 @@ void CreateICCCurvCurvTag(const std::vector<uint16_t>& curve,
   }
 }
 
+// Writes 12 + 4*params.size() bytes
 Status CreateICCCurvParaTag(std::vector<float> params, size_t curve_type,
                             PaddedBytes* JXL_RESTRICT tags) {
   WriteICCTag("para", tags->size(), tags);
@@ -360,6 +401,109 @@ Status CreateICCCurvParaTag(std::vector<float> params, size_t curve_type,
   }
   return true;
 }
+
+Status CreateICCLutAtoBTagForXYB(PaddedBytes* JXL_RESTRICT tags) {
+  WriteICCTag("mAB ", tags->size(), tags);
+  // 4 reserved bytes set to 0
+  WriteICCUint32(0, tags->size(), tags);
+  // number of input channels
+  WriteICCUint8(3, tags->size(), tags);
+  // number of output channels
+  WriteICCUint8(3, tags->size(), tags);
+  // 2 reserved bytes for padding
+  WriteICCUint16(0, tags->size(), tags);
+  // offset to first B curve
+  WriteICCUint32(32, tags->size(), tags);
+  // offset to matrix
+  WriteICCUint32(244, tags->size(), tags);
+  // offset to first M curve
+  WriteICCUint32(148, tags->size(), tags);
+  // offset to CLUT
+  WriteICCUint32(80, tags->size(), tags);
+  // offset to first A curve
+  // (reuse linear B curves)
+  WriteICCUint32(32, tags->size(), tags);
+
+  // offset = 32
+  // no-op curves
+  JXL_RETURN_IF_ERROR(CreateICCCurvParaTag({1.0f}, 0, tags));
+  JXL_RETURN_IF_ERROR(CreateICCCurvParaTag({1.0f}, 0, tags));
+  JXL_RETURN_IF_ERROR(CreateICCCurvParaTag({1.0f}, 0, tags));
+  // offset = 80
+  // number of grid points for each input channel
+  for (int i = 0; i < 16; ++i) {
+    WriteICCUint8(i < 3 ? 2 : 0, tags->size(), tags);
+  }
+  // precision = 2
+  WriteICCUint8(2, tags->size(), tags);
+  // 3 bytes of padding
+  WriteICCUint8(0, tags->size(), tags);
+  WriteICCUint16(0, tags->size(), tags);
+  const float kOffsets[3] = {
+      kScaledXYBOffset[0] + kScaledXYBOffset[1],
+      kScaledXYBOffset[1] - kScaledXYBOffset[0] + 1.0f / kScaledXYBScale[0],
+      kScaledXYBOffset[1] + kScaledXYBOffset[2]};
+  const float kScaling[3] = {
+      1.0f / (1.0f / kScaledXYBScale[0] + 1.0f / kScaledXYBScale[1]),
+      1.0f / (1.0f / kScaledXYBScale[0] + 1.0f / kScaledXYBScale[1]),
+      1.0f / (1.0f / kScaledXYBScale[1] + 1.0f / kScaledXYBScale[2])};
+  // 2*2*2*3 entries of 2 bytes each = 48 bytes
+  for (size_t ix = 0; ix < 2; ++ix) {
+    for (size_t iy = 0; iy < 2; ++iy) {
+      for (size_t ib = 0; ib < 2; ++ib) {
+        float in_f[3] = {ix * 1.0f, iy * 1.0f, ib * 1.0f};
+        for (size_t c = 0; c < 3; ++c) {
+          in_f[c] /= kScaledXYBScale[c];
+          in_f[c] -= kScaledXYBOffset[c];
+        }
+        float out_f[3];
+        out_f[0] = in_f[1] + in_f[0];
+        out_f[1] = in_f[1] - in_f[0];
+        out_f[2] = in_f[2] + in_f[1];
+        for (int i = 0; i < 3; ++i) {
+          out_f[i] += kOffsets[i];
+          out_f[i] *= kScaling[i];
+        }
+        for (int i = 0; i < 3; ++i) {
+          JXL_RETURN_IF_ERROR(out_f[i] >= 0.f && out_f[i] <= 1.f);
+          uint16_t val = static_cast<uint16_t>(
+              0.5f + 65535 * std::max(0.f, std::min(1.f, out_f[i])));
+          WriteICCUint16(val, tags->size(), tags);
+        }
+      }
+    }
+  }
+  // offset = 148
+  // 3 curves with 5 parameters = 3 * (12 + 5 * 4) = 96 bytes
+  for (size_t i = 0; i < 3; ++i) {
+    const float b =
+        -kOffsets[i] - std::cbrt(jxl::kNegOpsinAbsorbanceBiasRGB[i]);
+    std::vector<float> params = {
+        3,
+        1.0f / kScaling[i],
+        b,
+        0,                                // unused
+        std::max(0.f, -b * kScaling[i]),  // make skcms happy
+    };
+    JXL_RETURN_IF_ERROR(CreateICCCurvParaTag(params, 3, tags));
+  }
+  // offset = 244
+  const double matrix[] = {1.5170095, -1.1065225, 0.071623,
+                           -0.050022, 0.5683655,  -0.018344,
+                           -1.387676, 1.1145555,  0.6857255};
+  // 12 * 4 = 48 bytes
+  for (size_t i = 0; i < 9; ++i) {
+    JXL_RETURN_IF_ERROR(WriteICCS15Fixed16(matrix[i], tags->size(), tags));
+  }
+  for (size_t i = 0; i < 3; ++i) {
+    float intercept = 0;
+    for (size_t j = 0; j < 3; ++j) {
+      intercept += matrix[i * 3 + j] * jxl::kNegOpsinAbsorbanceBiasRGB[j];
+    }
+    JXL_RETURN_IF_ERROR(WriteICCS15Fixed16(intercept, tags->size(), tags));
+  }
+  return true;
+}
 }  // namespace
 
 Status MaybeCreateProfile(const ColorEncoding& c,
@@ -373,14 +517,20 @@ Status MaybeCreateProfile(const ColorEncoding& c,
   switch (c.GetColorSpace()) {
     case ColorSpace::kRGB:
     case ColorSpace::kGray:
-      break;  // OK
     case ColorSpace::kXYB:
-      return JXL_FAILURE("XYB ICC not yet implemented");
+      break;  // OK
     default:
       return JXL_FAILURE("Invalid CS %u",
                          static_cast<unsigned int>(c.GetColorSpace()));
   }
 
+  if (c.GetColorSpace() == ColorSpace::kXYB &&
+      c.rendering_intent != RenderingIntent::kPerceptual) {
+    return JXL_FAILURE(
+        "Only perceptual rendering intent implemented for XYB "
+        "ICC profile.");
+  }
+
   JXL_RETURN_IF_ERROR(CreateICCHeader(c, &header));
 
   std::vector<size_t> offsets;
@@ -393,9 +543,7 @@ Status MaybeCreateProfile(const ColorEncoding& c,
   FinalizeICCTag(&tags, &tag_offset, &tag_size);
   AddToICCTagTable("desc", tag_offset, tag_size, &tagtable, &offsets);
 
-  const std::string copyright =
-      "Copyright 2019 Google LLC, CC-BY-SA 3.0 Unported "
-      "license(https://creativecommons.org/licenses/by-sa/3.0/legalcode)";
+  const std::string copyright = "CC0";
   CreateICCMlucTag(copyright, &tags);
   FinalizeICCTag(&tags, &tag_offset, &tag_size);
   AddToICCTagTable("cprt", tag_offset, tag_size, &tagtable, &offsets);
@@ -417,6 +565,15 @@ Status MaybeCreateProfile(const ColorEncoding& c,
     float chad[9];
     JXL_RETURN_IF_ERROR(CreateICCChadMatrix(c.GetWhitePoint(), chad));
 
+    JXL_RETURN_IF_ERROR(CreateICCChadTag(chad, &tags));
+    FinalizeICCTag(&tags, &tag_offset, &tag_size);
+    AddToICCTagTable("chad", tag_offset, tag_size, &tagtable, &offsets);
+  }
+
+  if (c.GetColorSpace() == ColorSpace::kRGB) {
+    MaybeCreateICCCICPTag(c, &tags, &tag_offset, &tag_size, &tagtable,
+                          &offsets);
+
     const PrimariesCIExy primaries = c.GetPrimaries();
     float m[9];
     JXL_RETURN_IF_ERROR(CreateICCRGBMatrix(primaries.r, primaries.g,
@@ -425,10 +582,6 @@ Status MaybeCreateProfile(const ColorEncoding& c,
     float g[3] = {m[1], m[4], m[7]};
     float b[3] = {m[2], m[5], m[8]};
 
-    JXL_RETURN_IF_ERROR(CreateICCChadTag(chad, &tags));
-    FinalizeICCTag(&tags, &tag_offset, &tag_size);
-    AddToICCTagTable("chad", tag_offset, tag_size, &tagtable, &offsets);
-
     JXL_RETURN_IF_ERROR(CreateICCXYZTag(r, &tags));
     FinalizeICCTag(&tags, &tag_offset, &tag_size);
     AddToICCTagTable("rXYZ", tag_offset, tag_size, &tagtable, &offsets);
@@ -442,48 +595,58 @@ Status MaybeCreateProfile(const ColorEncoding& c,
     AddToICCTagTable("bXYZ", tag_offset, tag_size, &tagtable, &offsets);
   }
 
-  if (c.tf.IsGamma()) {
-    float gamma = 1.0 / c.tf.GetGamma();
-    JXL_RETURN_IF_ERROR(
-        CreateICCCurvParaTag({gamma, 1.0, 0.0, 1.0, 0.0}, 3, &tags));
+  if (c.GetColorSpace() == ColorSpace::kXYB) {
+    JXL_RETURN_IF_ERROR(CreateICCLutAtoBTagForXYB(&tags));
+    FinalizeICCTag(&tags, &tag_offset, &tag_size);
+    AddToICCTagTable("A2B0", tag_offset, tag_size, &tagtable, &offsets);
   } else {
-    switch (c.tf.GetTransferFunction()) {
-      case TransferFunction::kHLG:
-        CreateICCCurvCurvTag(
-            HWY_DYNAMIC_DISPATCH(CreateTableCurve)(4096, ExtraTF::kHLG), &tags);
-        break;
-      case TransferFunction::kPQ:
-        CreateICCCurvCurvTag(
-            HWY_DYNAMIC_DISPATCH(CreateTableCurve)(4096, ExtraTF::kPQ), &tags);
-        break;
-      case TransferFunction::kSRGB:
-        JXL_RETURN_IF_ERROR(CreateICCCurvParaTag(
-            {2.4, 1.0 / 1.055, 0.055 / 1.055, 1.0 / 12.92, 0.04045}, 3, &tags));
-        break;
-      case TransferFunction::k709:
-        JXL_RETURN_IF_ERROR(CreateICCCurvParaTag(
-            {1.0 / 0.45, 1.0 / 1.099, 0.099 / 1.099, 1.0 / 4.5, 0.081}, 3,
-            &tags));
-        break;
-      case TransferFunction::kLinear:
-        JXL_RETURN_IF_ERROR(
-            CreateICCCurvParaTag({1.0, 1.0, 0.0, 1.0, 0.0}, 3, &tags));
-        break;
-      case TransferFunction::kDCI:
-        JXL_RETURN_IF_ERROR(
-            CreateICCCurvParaTag({2.6, 1.0, 0.0, 1.0, 0.0}, 3, &tags));
-        break;
-      default:
-        JXL_ABORT("Unknown TF %d", c.tf.GetTransferFunction());
+    if (c.tf.IsGamma()) {
+      float gamma = 1.0 / c.tf.GetGamma();
+      JXL_RETURN_IF_ERROR(
+          CreateICCCurvParaTag({gamma, 1.0, 0.0, 1.0, 0.0}, 3, &tags));
+    } else if (c.GetColorSpace() != ColorSpace::kXYB) {
+      switch (c.tf.GetTransferFunction()) {
+        case TransferFunction::kHLG:
+          CreateICCCurvCurvTag(
+              HWY_DYNAMIC_DISPATCH(CreateTableCurve)(4096, ExtraTF::kHLG),
+              &tags);
+          break;
+        case TransferFunction::kPQ:
+          CreateICCCurvCurvTag(
+              HWY_DYNAMIC_DISPATCH(CreateTableCurve)(4096, ExtraTF::kPQ),
+              &tags);
+          break;
+        case TransferFunction::kSRGB:
+          JXL_RETURN_IF_ERROR(CreateICCCurvParaTag(
+              {2.4, 1.0 / 1.055, 0.055 / 1.055, 1.0 / 12.92, 0.04045}, 3,
+              &tags));
+          break;
+        case TransferFunction::k709:
+          JXL_RETURN_IF_ERROR(CreateICCCurvParaTag(
+              {1.0 / 0.45, 1.0 / 1.099, 0.099 / 1.099, 1.0 / 4.5, 0.081}, 3,
+              &tags));
+          break;
+        case TransferFunction::kLinear:
+          JXL_RETURN_IF_ERROR(
+              CreateICCCurvParaTag({1.0, 1.0, 0.0, 1.0, 0.0}, 3, &tags));
+          break;
+        case TransferFunction::kDCI:
+          JXL_RETURN_IF_ERROR(
+              CreateICCCurvParaTag({2.6, 1.0, 0.0, 1.0, 0.0}, 3, &tags));
+          break;
+        default:
+          JXL_ABORT("Unknown TF %u",
+                    static_cast<unsigned int>(c.tf.GetTransferFunction()));
+      }
+    }
+    FinalizeICCTag(&tags, &tag_offset, &tag_size);
+    if (c.IsGray()) {
+      AddToICCTagTable("kTRC", tag_offset, tag_size, &tagtable, &offsets);
+    } else {
+      AddToICCTagTable("rTRC", tag_offset, tag_size, &tagtable, &offsets);
+      AddToICCTagTable("gTRC", tag_offset, tag_size, &tagtable, &offsets);
+      AddToICCTagTable("bTRC", tag_offset, tag_size, &tagtable, &offsets);
     }
-  }
-  FinalizeICCTag(&tags, &tag_offset, &tag_size);
-  if (c.IsGray()) {
-    AddToICCTagTable("kTRC", tag_offset, tag_size, &tagtable, &offsets);
-  } else {
-    AddToICCTagTable("rTRC", tag_offset, tag_size, &tagtable, &offsets);
-    AddToICCTagTable("gTRC", tag_offset, tag_size, &tagtable, &offsets);
-    AddToICCTagTable("bTRC", tag_offset, tag_size, &tagtable, &offsets);
   }
 
   // Tag count
@@ -505,8 +668,10 @@ Status MaybeCreateProfile(const ColorEncoding& c,
   // TODO(lode): manually verify with a reliable tool that this creates correct
   // signature (profile id) for ICC profiles.
   PaddedBytes icc_sum = *icc;
-  memset(icc_sum.data() + 44, 0, 4);
-  memset(icc_sum.data() + 64, 0, 4);
+  if (icc_sum.size() >= 64 + 4) {
+    memset(icc_sum.data() + 44, 0, 4);
+    memset(icc_sum.data() + 64, 0, 4);
+  }
   uint8_t checksum[16];
   ICCComputeMD5(icc_sum, checksum);
 
